"""Decode GN front end features for Thermal Dataset

For ECCV Supplementary Material.

Paper ID: 2499
"""

from __future__ import print_function, absolute_import

import os
import argparse

import numpy as np
from skimage.io import imread
from skimage.transform import rescale
import torch
import torchvision.transforms as transforms
from torch.optim import Adam
from torch.nn import MSELoss

from evtrans.pyyolo.model import YoloNetV3
from evtrans.networks import EVYOLOFrontend
from evtrans.networks import FrameYOLOFrontend

import matplotlib.pyplot as plt


def visualize_feature(model, processed_img, hidden_feature,
                      iteration=1000):
    optimizer = Adam([processed_img], lr=0.01)

    conv_output = 0

    for i in range(1, iteration+1):
        optimizer.zero_grad()

        conv_output = model(processed_img)

        diff_i = torch.mean(torch.abs(
            processed_img[:, :, :, 1:]-processed_img[:, :, :, :-1]))
        diff_j = torch.mean(torch.abs(
            processed_img[:, :, 1:, :]-processed_img[:, :, :-1, :]))
        tv_loss = 5.*(diff_i + diff_j)

        loss = MSELoss()(conv_output, hidden_feature)+tv_loss
        print("Iteration: {} Loss: {} {}".format(i, loss, tv_loss), end="\r")

        loss.backward()

        optimizer.step()

    return processed_img


# device
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

#  get on a parser
parser = argparse.ArgumentParser()

parser.add_argument("--thermal_file", type=str)

# logging directory
parser.add_argument("--checkpoint", type=str)

# number of convolution input dimension
parser.add_argument("--conv_input_dim", type=int)

# network weight path
parser.add_argument("--weights_path", type=str,
                    default=os.path.join("..", "weights",
                                         "yolov3_original.pth"))
# cutoff layer
parser.add_argument("--cut_stage", type=int)

args = parser.parse_args()

# load pretrained model
pretrain_model = YoloNetV3(nms=True)
pretrain_model.load_state_dict(torch.load(args.weights_path))
pretrain_model.to(DEVICE)
pretrain_model.eval()

# check checkpoint
if os.path.isfile(args.checkpoint):
    checkpoint_path = args.checkpoint
elif os.path.isdir(args.checkpoint):
    # search for the best checkpoint
    log_path = os.path.join(args.checkpoint, "log.csv")
    exp_log = np.loadtxt(log_path, delimiter=",", skiprows=1,
                         dtype=float)
    val_log = exp_log[:, 8]
    checkpoint_idx = np.argmin(val_log)+1
    checkpoint_path = os.path.join(
        args.checkpoint, "checkpoints",
        "checkpoint_{}.pt".format(int(checkpoint_idx)))
    print("Using checkpoint {}".format(checkpoint_path))

checkpoint = torch.load(checkpoint_path)
print("Checkpoint loaded.")

# build Event driven front end
ev_yolo = EVYOLOFrontend(stage=args.cut_stage,
                         input_dim=args.conv_input_dim).to(DEVICE)
ev_yolo.load_state_dict(checkpoint["model_state_dict"])
ev_yolo.eval()
print("Parameter loaded.")

yolo_frontend = FrameYOLOFrontend(
    pretrain_model.darknet, stage=args.cut_stage).to(DEVICE)
yolo_frontend.eval()


# load thermal image
thermal_img = imread(args.thermal_file)
thermal_img = transforms.ToTensor()(thermal_img)

# get image shape
img_shape = thermal_img.shape[1:3]

rgb_file = args.thermal_file.replace("thermal_8_bit", "RGB")
rgb_file = rgb_file.replace("jpeg", "jpg")
rgb_img = imread(rgb_file)
rgb_img = rescale(rgb_img, 0.4, multichannel=True)
rgb_img = rgb_img[
    58:58+thermal_img.shape[1], 65:65+thermal_img.shape[2], :]

candidate_ev_vol = thermal_img

# target random image
random_rgb_img = np.uint8(np.random.uniform(
    128-64, 128+64, (thermal_img.shape[1], thermal_img.shape[2], 3)))
processed_rgb_img = transforms.ToTensor()(random_rgb_img)
processed_rgb_img = processed_rgb_img.unsqueeze(0).to(DEVICE)
processed_rgb_img = processed_rgb_img.requires_grad_(True)

# inference on the data
with torch.no_grad():
    candidate = candidate_ev_vol.unsqueeze(0).to(DEVICE)
    pre_detections = ev_yolo(candidate)

rgb_feature = visualize_feature(yolo_frontend, processed_rgb_img,
                                pre_detections.detach(),
                                iteration=1000)

rgb_feature = rgb_feature.cpu().data.numpy()
rgb_feature = rgb_feature[0].transpose(1, 2, 0)
robust_max = np.percentile(rgb_feature.flatten(), 99)
robust_min = np.percentile(rgb_feature.flatten(), 1)
rgb_feature = np.clip(rgb_feature, robust_min, robust_max)
rgb_max = rgb_feature.max()
rgb_min = rgb_feature.min()
rgb_feature = (rgb_feature-rgb_min)/(rgb_max-rgb_min)

plt.figure()
plt.subplot(131)
plt.imshow(rgb_feature)
plt.axis("off")
plt.subplot(132)
thermal_img = torch.cat([thermal_img, thermal_img, thermal_img], dim=0)
plt.imshow(thermal_img.permute(1, 2, 0))
plt.axis("off")
plt.subplot(133)
plt.imshow(rgb_img)
plt.axis("off")
plt.show()
