"""NGA Loss function.

For ECCV Supplementary Material.

Paper ID: 2499
"""

from __future__ import print_function, absolute_import

from torch.nn import functional as F


def gram(x):
    (bs, ch, h, w) = x.shape
    m = x.mean().detach()
    f = x.flatten(start_dim=2)-m
    f_T = f.transpose(1, 2)
    G = f.bmm(f_T) / (ch * h * w)
    return G


def NGA_loss(event_batch, img_batch,
             ev_frontend, net_frontend, mid_net,
             frl=1., fel=1.,
             frl_gram=1e5, fel_gram=1e7,
             return_all_parts=False):
    """Calculate NGA Loss."""
    event_out = ev_frontend(event_batch)
    img_out = net_frontend(img_batch).detach()

    # Feature Reconstruction Loss
    frl_loss = frl*F.mse_loss(event_out, img_out.detach()) if frl != 0 else 0.

    # Fature Style Loss
    frl_gram_loss = frl_gram*F.mse_loss(
        gram(event_out), gram(img_out.detach())) if frl_gram != 0 else 0.

    # Feature Eavluation Loss
    _, ev_feature, img_feature = mid_net(
                event_out, img_out, use_last_only=True, return_feature=True)
    fel_loss = fel*F.mse_loss(
        ev_feature, img_feature.detach()) if fel != 0 else 0.

    # Evaluation style loss
    fel_gram_loss = fel_gram*F.mse_loss(
        gram(ev_feature),
        gram(img_feature.detach())) if fel_gram != 0 else 0.

    if return_all_parts is True:
        return_list = [frl_loss, fel_loss, frl_gram_loss, fel_gram_loss,
                       frl_loss+frl_gram_loss+fel_loss+fel_gram_loss,
                       F.adaptive_avg_pool2d(ev_feature, 1).squeeze(),
                       F.adaptive_avg_pool2d(img_feature, 1).squeeze()]
    else:
        return_list = [frl_loss+frl_gram_loss+fel_loss+fel_gram_loss,
                       F.adaptive_avg_pool2d(ev_feature, 1).squeeze(),
                       F.adaptive_avg_pool2d(img_feature, 1).squeeze()]

    return tuple(return_list)
