"""Training script for the thermal driven YOLO.

For ECCV Supplementary Material.

Paper ID: 2499
"""

from __future__ import print_function, absolute_import

import argparse
import os

import numpy as np

import torch
from torch.utils.data import DataLoader
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau

from evtrans.pyyolo.model import YoloNetV3
from evtrans.networks import EVYOLOFrontend
from evtrans.networks import FrameYOLOFrontend
from evtrans.networks import EVYOLOEvalNet
from evtrans.data import ThermalYOLODataset
from evtrans.loss import NGA_loss
from evtrans import log_history

# device
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"


#  get on a parser
parser = argparse.ArgumentParser()

# image size
parser.add_argument("--img_size", type=int, default=416)
# number of epochs
parser.add_argument("--num_epochs", type=int, default=10)
# batch size
parser.add_argument("--batch_size", type=int, default=4)
# training folder
parser.add_argument("--train_data_dir", type=str)
# validation folder
parser.add_argument("--val_data_dir", type=str)
# logging directory
parser.add_argument("--log_dir", type=str)

# network weight path
parser.add_argument("--weights_path", type=str,
                    default=os.path.join("..", "weights",
                                         "yolov3_original.pth"))
# cutoff layer
parser.add_argument("--cut_stage", type=int)
# eval end layer
parser.add_argument("--eval_stage", type=int)

# number of convolution input dimension
parser.add_argument("--conv_input_dim", type=int)
parser.add_argument("--num_eval_inputs", type=int)

# loss term coefficient
parser.add_argument("--frl", type=float, default=0.)
parser.add_argument("--fel", type=float, default=0.)
parser.add_argument("--frl_gram", type=float, default=0.)
parser.add_argument("--fel_gram", type=float, default=0.)

# parse argument
args = parser.parse_args()

# set up logging server
checkpoint_dir = os.path.join(args.log_dir, "checkpoints")
if not os.path.isdir(checkpoint_dir):
    os.makedirs(checkpoint_dir)
    print("Setup the logging directory at {}".format(args.log_dir))

log_file = os.path.join(args.log_dir, "log.csv")

# train and val data loader
train_dl = DataLoader(
    ThermalYOLODataset(args.train_data_dir, img_size=args.img_size,
                       is_train=True, parsing="/*.npz", no_resampling=True),
    batch_size=args.batch_size,
    shuffle=True,
    num_workers=4,
    drop_last=False)

val_dl = DataLoader(
    ThermalYOLODataset(args.val_data_dir, img_size=args.img_size,
                       is_train=False, parsing="/*.npz", no_resampling=True),
    batch_size=args.batch_size,
    shuffle=False,
    num_workers=4,
    drop_last=False)


# load pretrained model
pretrain_model = YoloNetV3(nms=True)
pretrain_model.load_state_dict(torch.load(args.weights_path))
pretrain_model.to(DEVICE)
pretrain_model.eval()

# build Event driven front end
thermal_yolo = EVYOLOFrontend(stage=args.cut_stage,
                              input_dim=args.conv_input_dim).to(DEVICE)

# made sure YOLO frontend and evaluation are at evaluation mode
yolo_frontend = FrameYOLOFrontend(
    pretrain_model.darknet, stage=args.cut_stage).to(DEVICE)
yolo_frontend.eval()

print("Initialize Thermal Frontend and Frontend Net")

yolo_eval_net = EVYOLOEvalNet(
    pretrain_model.darknet,
    stages=[args.cut_stage+1, args.eval_stage]).to(DEVICE)
yolo_eval_net.eval()

print("Initialized Middle Net")

optimizerG = Adam(thermal_yolo.parameters(), lr=1e-4)
scheduler = ReduceLROnPlateau(optimizerG, "min", factor=0.1, verbose=True)

num_batches = len(train_dl)

for epoch_i in range(args.num_epochs):
    print("*"*50)
    print("TRAINING EPOCH {}".format(epoch_i+1))

    epoch_losses = []
    epoch_frl = []
    epoch_fel = []
    epoch_frl_gram = []
    epoch_fel_gram = []
    for batch_i, (thermal_batch, img_batch) in enumerate(train_dl):
        thermal_yolo.train()

        # calculate generator loss and prepare features
        loss_output = NGA_loss(
            thermal_batch.to(DEVICE), img_batch.to(DEVICE),
            thermal_yolo, yolo_frontend, yolo_eval_net,
            frl=args.frl, fel=args.fel,
            frl_gram=args.frl_gram, fel_gram=args.fel_gram,
            return_all_parts=True)

        (b_frl, b_fel, b_frl_gram, b_fel_gram,
         batch_loss, ev_feature, img_feature) = loss_output

        optimizerG.zero_grad()
        batch_loss.backward()

        optimizerG.step()

        # collecting log
        print("[TRAIN] L at batch {}: {:.3f} FRL: {:.3f} "
              "FEL: {:.3f} FRLG: {:.3f} "
              "FELG: {:.3f}".format(
                  batch_i, batch_loss,
                  b_frl, b_fel, b_frl_gram,
                  b_fel_gram), end='\r')

        epoch_losses.append(float(batch_loss.item()))
        epoch_frl.append(float(b_frl))
        epoch_fel.append(float(b_fel))
        epoch_frl_gram.append(float(b_frl_gram))
        epoch_fel_gram.append(float(b_fel_gram))

    print("\n"+"-"*50)
    # record training loss
    train_epoch_loss = np.mean(epoch_losses)
    train_frl = np.mean(epoch_frl)
    train_fel = np.mean(epoch_fel)
    train_frl_gram = np.mean(epoch_frl_gram)
    train_fel_gram = np.mean(epoch_fel_gram)

    print("[TRAIN] L at epoch {}: {:.3f} FRL: {:.3f} "
          "FEL: {:.3f} FRLG: {:.3f} "
          "FELG: {:.3f}".format(
              epoch_i+1, train_epoch_loss, train_frl, train_fel,
              train_frl_gram, train_fel_gram))
    print("\n"+"-"*50)

    # evaluation pipeline
    epoch_losses = []
    epoch_frl = []
    epoch_fel = []
    epoch_frl_gram = []
    epoch_fel_gram = []

    for batch_i, (thermal_batch, img_batch) in enumerate(val_dl):
        thermal_yolo.eval()

        # calculate loss
        with torch.no_grad():
            # calculate loss
            loss_output = NGA_loss(
                thermal_batch.to(DEVICE), img_batch.to(DEVICE),
                thermal_yolo, yolo_frontend, yolo_eval_net,
                frl=args.frl, fel=args.fel,
                frl_gram=args.frl_gram, fel_gram=args.fel_gram,
                return_all_parts=True)

            # decode batch output
            (b_frl, b_fel, b_frl_gram, b_fel_gram,
             batch_loss, _, _) = loss_output

            # collect log
            epoch_losses.append(float(batch_loss.item()))
            epoch_frl.append(float(b_frl))
            epoch_fel.append(float(b_fel))
            epoch_frl_gram.append(float(b_frl_gram))
            epoch_fel_gram.append(float(b_fel_gram))

            print("[VALID] L at batch {}: {:.3f} FRL: {:.3f} "
                  "FEL: {:.3f} FRLG: {:.3f} "
                  "FELG: {:.3f}".format(
                      batch_i, batch_loss,
                      b_frl, b_fel, b_frl_gram,
                      b_fel_gram), end='\r')

    # calculate validation loss
    val_epoch_loss = np.mean(epoch_losses)
    val_frl = np.mean(epoch_frl)
    val_fel = np.mean(epoch_fel)
    val_frl_gram = np.mean(epoch_frl_gram)
    val_fel_gram = np.mean(epoch_fel_gram)
    print("-"*50)
    print("[VALID] L at epoch {}: {:.3f} FRL: {:.3f} "
          "FEL: {:.3f} FRLG: {:.3f} "
          "FELG: {:.3f}".format(
              epoch_i+1, val_epoch_loss, val_frl, val_fel,
              val_frl_gram, val_fel_gram))

    # scheduling the loss
    scheduler.step(val_epoch_loss)

    # write log into file
    log_history(log_file,
                [train_epoch_loss,
                 train_frl, train_fel, train_frl_gram,
                 train_fel_gram,
                 val_epoch_loss,
                 val_frl, val_fel, val_frl_gram,
                 val_fel_gram],
                idx=epoch_i+1,
                header=["epoch",
                        "train_loss",
                        "train_frl", "train_fel",
                        "train_frl_gram", "train_fel_gram",
                        "val_loss",
                        "val_frl", "val_fel",
                        "val_frl_gram", "val_fel_gram"])

    print("*"*50)

    # Saving the network
    checkpoint_path = os.path.join(
        checkpoint_dir, "checkpoint_{}.pt".format(epoch_i+1))
    torch.save(
        {"epoch": epoch_i,
         "model_state_dict": thermal_yolo.state_dict()},
        checkpoint_path)
